@@ -1,9 +1,6 @@
 /***************************************************************************
- $RCSfile$
-                             -------------------
-    cvs         : $Id$
     begin       : Mon Mar 01 2004
-    copyright   : (C) 2004 by Martin Preuss
+    copyright   : (C) 2004-2011 by Martin Preuss
     email       : martin@libchipcard.de
 
  ***************************************************************************
@@ -25,11 +22,17 @@
 #include <aqhbci/provider.h>
 #include "adminjobs_l.h"
 
+#include "hhd_l.h"
+
+#include <aqbanking/banking_be.h>
+
 #include <gwenhywfar/debug.h>
+#include <gwenhywfar/text.h>
 
 #include <stdlib.h>
 #include <assert.h>
 #include <string.h>
+#include <ctype.h>
 
 
 GWEN_INHERIT(AB_USER, AH_USER)
@@ -83,6 +86,15 @@ void AH_User_Flags_toDb(GWEN_DB_NODE *db, const char *name,
   if (flags & AH_USER_FLAGS_NO_BASE64)
     GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_DEFAULT, name,
 			 "noBase64");
+  if (flags & AH_USER_FLAGS_KEEP_MULTIPLE_BLANKS)
+    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_DEFAULT, name,
+			 "keepMultipleBlanks");
+  if (flags & AH_USER_FLAGS_TAN_OMIT_SMS_ACCOUNT)
+    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_DEFAULT, name,
+                         "omitSmsAccount");
+  if (flags & AH_USER_FLAGS_TLS_ONLY_SAFE_CIPHERS)
+    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_DEFAULT, name,
+                         "tlsOnlySafeCiphers");
 }
 
 
@@ -105,10 +117,17 @@ uint32_t AH_User_Flags_fromDb(GWEN_DB_NODE *db, const char *name) {
       fl|=AH_USER_FLAGS_KEEPALIVE;
     else if (strcasecmp(s, "ignoreUpd")==0)
       fl|=AH_USER_FLAGS_IGNORE_UPD;
-    else if (strcasecmp(s, "forceSsl3")==0)
+    else if (strcasecmp(s, "forceSsl3")==0 ||
+	     strcasecmp(s, "forceSslv3")==0)
       fl|=AH_USER_FLAGS_FORCE_SSL3;
     else if (strcasecmp(s, "noBase64")==0)
       fl|=AH_USER_FLAGS_NO_BASE64;
+    else if (strcasecmp(s, "keepMultipleBlanks")==0)
+      fl|=AH_USER_FLAGS_KEEP_MULTIPLE_BLANKS;
+    else if (strcasecmp(s, "omitSmsAccount")==0)
+      fl|=AH_USER_FLAGS_TAN_OMIT_SMS_ACCOUNT;
+    else if (strcasecmp(s, "tlsOnlySafeCiphers")==0)
+      fl|=AH_USER_FLAGS_TLS_ONLY_SAFE_CIPHERS;
     else {
       DBG_WARN(AQHBCI_LOGDOMAIN, "Unknown user flag \"%s\"", s);
     }
@@ -121,9 +140,10 @@ uint32_t AH_User_Flags_fromDb(GWEN_DB_NODE *db, const char *name) {
 
 
 
-void AH_User_Extend(AB_USER *u, AB_PROVIDER *pro,
-		    AB_PROVIDER_EXTEND_MODE em,
-		    GWEN_DB_NODE *db) {
+int AH_User_Extend(AB_USER *u, AB_PROVIDER *pro,
+                   AB_PROVIDER_EXTEND_MODE em,
+                   GWEN_DB_NODE *db) {
+  DBG_INFO(AQHBCI_LOGDOMAIN, "Extending user with mode %d", em);
   if (em==AB_ProviderExtendMode_Create ||
       em==AB_ProviderExtendMode_Extend) {
     AH_USER *ue;
@@ -138,6 +158,7 @@ void AH_User_Extend(AB_USER *u, AB_PROVIDER *pro,
 
     ue->hbci=AH_Provider_GetHbci(pro);
     ue->tanMethodDescriptions=AH_TanMethod_List_new();
+    ue->sepaDescriptors=GWEN_StringList_new();
 
     s=AB_User_GetCountry(u);
     if (!s || !*s)
@@ -155,18 +176,40 @@ void AH_User_Extend(AB_USER *u, AB_PROVIDER *pro,
       ue->hbciVersion=210;
       ue->bpd=AH_Bpd_new();
       ue->dbUpd=GWEN_DB_Group_new("upd");
+      ue->maxTransfersPerJob=AH_USER_MAX_TRANSFERS_PER_JOB;
+      ue->maxDebitNotesPerJob=AH_USER_MAX_DEBITNOTES_PER_JOB;
     }
     else {
       /* update db to latest version */
       rv=AH_HBCI_UpdateDbUser(ue->hbci, db);
-      if (rv) {
+      if (rv<0) {
 	DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not update user db (%d)", rv);
-	assert(0);
+        GWEN_Gui_ShowError(I18N("AqBanking Settings Database Error"),
+                           I18N("Your settings database might be in an inconsistent state!"));
+        return rv;
       }
       AH_User_ReadDb(u, db);
       AH_User_LoadTanMethods(u);
+      AH_User_LoadSepaDescriptors(u);
+      if (rv==1) {
+	/* updated config, write it now */
+        DBG_NOTICE(AQHBCI_LOGDOMAIN, "Writing back updated HBCI user %d", AB_User_GetUniqueId(u));
+	rv=AB_Banking_SaveUserConfig(AB_Provider_GetBanking(pro), u, 1);
+	if (rv<0) {
+          DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not save user db (%d)", rv);
+          GWEN_Gui_ShowError(I18N("AqBanking Settings Database Error"),
+                             I18N("Your settings database might be in an inconsistent state!"));
+          return rv;
+	}
+      }
     }
   }
+  else if (em==AB_ProviderExtendMode_Reload) {
+    /* just reload user */
+    AH_User_ReadDb(u, db);
+    AH_User_LoadTanMethods(u);
+    AH_User_LoadSepaDescriptors(u);
+  }
   else {
     AH_USER *ue;
 
@@ -179,6 +222,8 @@ void AH_User_Extend(AB_USER *u, AB_PROVIDER *pro,
       AH_User_toDb(u, db);
     } /* if save */
   }
+
+  return 0;
 }
 
 
@@ -194,9 +239,11 @@ void GWENHYWFAR_CB AH_User_freeData(void *bp, void *p) {
   free(ue->tokenName);
   free(ue->prompt);
   GWEN_Url_free(ue->serverUrl);
+  GWEN_DB_Group_free(ue->dbUpd);
   AH_Bpd_free(ue->bpd);
   GWEN_MsgEngine_free(ue->msgEngine);
   AH_TanMethod_List_free(ue->tanMethodDescriptions);
+  GWEN_StringList_free(ue->sepaDescriptors);
   GWEN_FREE_OBJECT(ue);
 }
 
@@ -239,7 +286,7 @@ void AH_User_ReadDb(AB_USER *u, GWEN_DB_NODE *db) {
   }
   else
     ue->serverUrl=NULL;
-  
+
   /* load BPD */
   AH_Bpd_free(ue->bpd);
   gr=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "bpd");
@@ -254,8 +301,9 @@ void AH_User_ReadDb(AB_USER *u, GWEN_DB_NODE *db) {
   if (ue->dbUpd)
     GWEN_DB_Group_free(ue->dbUpd);
   gr=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "upd");
-  if (gr)
+  if (gr) {
     ue->dbUpd=GWEN_DB_Group_dup(gr);
+  }
   else
     ue->dbUpd=GWEN_DB_Group_new("upd");
   
@@ -304,8 +352,14 @@ void AH_User_ReadDb(AB_USER *u, GWEN_DB_NODE *db) {
   
   /* get rdh type */
   ue->rdhType=GWEN_DB_GetIntValue(db, "rdhType", 0, -1);
-  
+  if (ue->rdhType<1)
+    ue->rdhType=1;
+
   /* read supported TAN methods */
+  for (i=0; i<AH_USER_MAX_TANMETHODS; i++)
+    ue->tanMethodList[i]=-1;
+  ue->tanMethodCount=0;
+
   for (i=0; i<AH_USER_MAX_TANMETHODS; i++) {
     int method;
   
@@ -317,6 +371,23 @@ void AH_User_ReadDb(AB_USER *u, GWEN_DB_NODE *db) {
   }
   
   ue->selectedTanMethod=GWEN_DB_GetIntValue(db, "selectedTanMethod", 0, 0);
+
+  /* read some settings */
+  ue->maxTransfersPerJob=GWEN_DB_GetIntValue(db, "maxTransfersPerJob", 0, AH_USER_MAX_TRANSFERS_PER_JOB);
+  ue->maxDebitNotesPerJob=GWEN_DB_GetIntValue(db, "maxDebitNotesPerJob", 0, AH_USER_MAX_DEBITNOTES_PER_JOB);
+  free(ue->sepaTransferProfile);
+  s=GWEN_DB_GetCharValue(db, "sepaTransferProfile", 0, NULL);
+  if (s) ue->sepaTransferProfile=strdup(s);
+  else ue->sepaTransferProfile=NULL;
+  free(ue->sepaDebitNoteProfile);
+  s=GWEN_DB_GetCharValue(db, "sepaDebitNoteProfile", 0, NULL);
+  if (s) ue->sepaDebitNoteProfile=strdup(s);
+  else ue->sepaDebitNoteProfile=NULL;
+
+  free(ue->tanMediumId);
+  s=GWEN_DB_GetCharValue(db, "tanMediumId", 0, NULL);
+  if (s) ue->tanMediumId=strdup(s);
+  else ue->tanMediumId=NULL;
 }
 
 
@@ -423,9 +494,26 @@ void AH_User_toDb(AB_USER *u, GWEN_DB_NODE *db) {
 			ue->tanMethodList[i]);
   }
   
-  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_DEFAULT,
+  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
 		      "selectedTanMethod",
 		      ue->selectedTanMethod);
+
+  /* store some settings */
+  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
+		      "maxTransfersPerJob",
+		      ue->maxTransfersPerJob);
+  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
+		      "maxDebitNotesPerJob",
+		      ue->maxDebitNotesPerJob);
+  if (ue->sepaTransferProfile)
+    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
+			 "sepaTransferProfile", ue->sepaTransferProfile);
+  if (ue->sepaDebitNoteProfile)
+    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
+			 "sepaDebitNoteProfile", ue->sepaDebitNoteProfile);
+  if (ue->tanMediumId)
+    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
+			 "tanMediumId", ue->tanMediumId);
 }
 
 
@@ -604,6 +692,58 @@ GWEN_DB_NODE *AH_User_GetUpd(const AB_USER *u){
 
 
 
+GWEN_DB_NODE *AH_User_GetUpdForAccountIdAndSuffix(const AB_USER *u,
+                                                  const char *sAccountNumber,
+                                                  const char *sAccountSuffix) {
+  AH_USER *ue;
+  GWEN_DB_NODE *db;
+  GWEN_BUFFER *tbuf;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  db=AH_User_GetUpd(u);
+  if (db==NULL) {
+    DBG_INFO(AQHBCI_LOGDOMAIN, "No upd");
+    return NULL;
+  }
+
+  tbuf=GWEN_Buffer_new(0, 64, 0, 1);
+  GWEN_Buffer_AppendString(tbuf, sAccountNumber);
+  GWEN_Buffer_AppendString(tbuf, "-");
+  /* take into account the "Unterkontomerkmal", don't rely solely on account id */
+  if (sAccountSuffix && *sAccountSuffix)
+    GWEN_Buffer_AppendString(tbuf, sAccountSuffix);
+  else
+    GWEN_Buffer_AppendString(tbuf, "none");
+  DBG_INFO(AQHBCI_LOGDOMAIN, "Checking upd for account \"%s\"", GWEN_Buffer_GetStart(tbuf));
+  db=GWEN_DB_GetGroup(db,
+                      GWEN_PATH_FLAGS_NAMEMUSTEXIST,
+                      GWEN_Buffer_GetStart(tbuf));
+  GWEN_Buffer_free(tbuf);
+
+  if (db==NULL) {
+    DBG_INFO(AQHBCI_LOGDOMAIN, "Falling back to old storage of UPD for account \"%s\"", sAccountNumber);
+    db=AH_User_GetUpd(u);
+    db=GWEN_DB_GetGroup(db,
+                        GWEN_PATH_FLAGS_NAMEMUSTEXIST,
+                        sAccountNumber);
+  }
+
+  return db;
+}
+
+
+
+GWEN_DB_NODE *AH_User_GetUpdForAccount(const AB_USER *u, const AB_ACCOUNT *acc){
+  return AH_User_GetUpdForAccountIdAndSuffix(u,
+                                             AB_Account_GetAccountNumber(acc),
+                                             AB_Account_GetSubAccountId(acc));
+}
+
+
+
 void AH_User_SetUpd(AB_USER *u, GWEN_DB_NODE *n){
   AH_USER *ue;
 
@@ -613,8 +753,9 @@ void AH_User_SetUpd(AB_USER *u, GWEN_DB_NODE *n){
 
   if (ue->dbUpd)
     GWEN_DB_Group_free(ue->dbUpd);
-  if (n)
+  if (n) {
     ue->dbUpd=GWEN_DB_Group_dup(n);
+  }
   else
     ue->dbUpd=GWEN_DB_Group_new("upd");
 }
@@ -1058,8 +1199,20 @@ int AH_User_MkTanName(const AB_USER *u,
     return 0;
   }
   else {
-    DBG_ERROR(AQHBCI_LOGDOMAIN, "Missing tokenType or tokenName");
-    return GWEN_ERROR_NO_DATA;
+    const char *s;
+
+    DBG_DEBUG(AQHBCI_LOGDOMAIN, "No tokenType or tokenName");
+    GWEN_Buffer_AppendString(buf, "TAN_");
+    s=AB_User_GetBankCode(u);
+    if (s)
+      GWEN_Buffer_AppendString(buf, s);
+    GWEN_Buffer_AppendString(buf, "_");
+    GWEN_Buffer_AppendString(buf, AB_User_GetUserId(u));
+    if (challenge) {
+      GWEN_Buffer_AppendString(buf, "_");
+      GWEN_Buffer_AppendString(buf, challenge);
+    }
+    return 0;
   }
 }
 
@@ -1068,7 +1221,7 @@ int AH_User_MkTanName(const AB_USER *u,
 int AH_User_InputPin(AB_USER *u,
 		     char *pwbuffer,
 		     int minLen, int maxLen,
-		     int flags){
+                     int flags){
   GWEN_BUFFER *nbuf;
   int rv;
   const char *numeric_warning = "";
@@ -1156,7 +1309,7 @@ int AH_User_InputPin(AB_USER *u,
 int AH_User_InputPasswd(AB_USER *u,
 			char *pwbuffer,
 			int minLen, int maxLen,
-			int flags){
+                        int flags){
   GWEN_BUFFER *nbuf;
   int rv;
   const char *numeric_warning = "";
@@ -1244,7 +1397,7 @@ int AH_User_InputPasswd(AB_USER *u,
 int AH_User_InputTan(AB_USER *u,
 		     char *pwbuffer,
 		     int minLen,
-		     int maxLen){
+                     int maxLen){
   int rv;
   char buffer[512];
   const char *un;
@@ -1301,12 +1454,13 @@ int AH_User_InputTanWithChallenge(AB_USER *u,
 				  const char *challenge,
 				  char *pwbuffer,
 				  int minLen,
-				  int maxLen){
+                                  int maxLen){
   int rv;
-  char buffer[512];
+  char buffer[1024];
   const char *un;
   const char *bn=NULL;
   GWEN_BUFFER *nbuf;
+  GWEN_BUFFER *xbuf;
   AB_BANKINFO *bi;
 
   assert(u);
@@ -1323,38 +1477,56 @@ int AH_User_InputTanWithChallenge(AB_USER *u,
 
   buffer[0]=0;
   buffer[sizeof(buffer)-1]=0;
+
+  xbuf=GWEN_Buffer_new(0, 256, 0, 1);
+
+  /* text version */
   snprintf(buffer, sizeof(buffer)-1,
-	   I18N("Please enter the TAN\n"
-		"for user %s at %s.\n"
-                "The server provided the following challenge:\n"
-                "%s"
-                "<html>"
-                "<p>"
-		"Please enter the TAN for user <i>%s</i> at "
-		"<i>%s</i>."
-		"</p>"
-                "<p>"
-                "The server provided the following challenge:"
-                "</p>"
-                "<p align=\"center\" >"
-                "<font color=\"blue\">%s</font>"
-                "</p>"
-                "</html>"),
-	   un, bn, challenge,
-	   un, bn, challenge);
+           I18N("Please enter the TAN\n"
+                "for user %s at %s.\n"), un, bn);
+  buffer[sizeof(buffer)-1]=0;
+  GWEN_Buffer_AppendString(xbuf, buffer);
+  if (challenge && *challenge) {
+    GWEN_Buffer_AppendString(xbuf, I18N("The server provided the following challenge:"));
+    GWEN_Buffer_AppendString(xbuf, "\n");
+    GWEN_Buffer_AppendString(xbuf, challenge);
+  }
+
+  /* html version */
+  GWEN_Buffer_AppendString(xbuf,
+                           "<html>"
+                           "<p>");
+  snprintf(buffer, sizeof(buffer)-1,
+           I18N("Please enter the TAN for user <i>%s</i> at <i>%s</i>."), un, bn);
+  buffer[sizeof(buffer)-1]=0;
+  GWEN_Buffer_AppendString(xbuf, buffer);
+  GWEN_Buffer_AppendString(xbuf, "</p>");
+  if (challenge && *challenge) {
+    GWEN_Buffer_AppendString(xbuf, "<p>");
+    GWEN_Buffer_AppendString(xbuf, I18N("The server provided the following challenge:"));
+    GWEN_Buffer_AppendString(xbuf, "</p>"
+                             "<p align=\"center\" >"
+                             "<font color=\"blue\">");
+    GWEN_Buffer_AppendString(xbuf, challenge);
+    GWEN_Buffer_AppendString(xbuf,
+                             "</font>"
+                             "</p>"
+                             "</html>");
+  }
 
   nbuf=GWEN_Buffer_new(0, 256 ,0 ,1);
   AH_User_MkTanName(u, challenge, nbuf);
   rv=GWEN_Gui_GetPassword(GWEN_GUI_INPUT_FLAGS_TAN |
 			  /*GWEN_GUI_INPUT_FLAGS_NUMERIC |*/
-			  GWEN_GUI_INPUT_FLAGS_SHOW,
+                          GWEN_GUI_INPUT_FLAGS_SHOW,
 			  GWEN_Buffer_GetStart(nbuf),
-			  I18N("Enter TAN"),
-			  buffer,
-			  pwbuffer,
+                          I18N("Enter TAN"),
+                          GWEN_Buffer_GetStart(xbuf),
+                          pwbuffer,
 			  minLen,
 			  maxLen,
 			  0);
+  GWEN_Buffer_free(xbuf);
   GWEN_Buffer_free(nbuf);
   AB_BankInfo_free(bi);
   return rv;
@@ -1452,6 +1624,7 @@ int AH_User_HasTanMethod(const AB_USER *u, int method) {
       return 1;
   }
 
+  DBG_INFO(AQHBCI_LOGDOMAIN, "TAN method %d not available", method);
   return 0;
 }
 
@@ -1541,63 +1714,497 @@ void AH_User_SetSelectedTanMethod(AB_USER *u, int i) {
 
 
 
-void AH_User_LoadTanMethods(AB_USER *u) {
+const char *AH_User_GetTanMediumId(const AB_USER *u) {
   AH_USER *ue;
-  AH_JOB *jTan;
 
   assert(u);
   ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
   assert(ue);
 
-  jTan=AH_Job_Tan_new(u, 1);
-  if (!jTan) {
-    DBG_WARN(AQHBCI_LOGDOMAIN, "Job HKTAN not available");
-  }
-  else {
+  return ue->tanMediumId;
+}
+
+
+
+void AH_User_SetTanMediumId(AB_USER *u, const char *s) {
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  free(ue->tanMediumId);
+  if (s) ue->tanMediumId=strdup(s);
+  else ue->tanMediumId=NULL;
+}
+
+
+
+void AH_User_LoadTanMethods(AB_USER *u) {
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  /* read directly from BPD */
+  if (ue->cryptMode==AH_CryptMode_Pintan) {
     GWEN_DB_NODE *db;
-    GWEN_DB_NODE *dbT;
+    int rv;
 
-    db=AH_Job_GetParams(jTan);
-    assert(db);
-
-    dbT=GWEN_DB_FindFirstGroup(db, "tanMethod");
-    while(dbT) {
-      AH_TAN_METHOD *tm;
-
-      tm=AH_TanMethod_fromDb(dbT);
-      if (tm)
-	AH_TanMethod_List_Add(tm, ue->tanMethodDescriptions);
-      else {
-	DBG_WARN(AQHBCI_LOGDOMAIN, "Invalid TAN method");
-	GWEN_DB_Dump(dbT, stderr, 2);
-      }
-
-      dbT=GWEN_DB_FindNextGroup(dbT, "tanMethod");
+    AH_TanMethod_List_Clear(ue->tanMethodDescriptions);
+    db=GWEN_DB_Group_new("bpd");
+    rv=AH_Job_SampleBpdVersions("JobTan", u, db);
+    if (rv<0) {
+      DBG_INFO(AQHBCI_LOGDOMAIN, "No BPD for TAN job");
     }
+    else {
+      GWEN_DB_NODE *dbV;
 
-    AH_Job_free(jTan);
+      dbV=GWEN_DB_GetFirstGroup(db);
+      while(dbV) {
+	int version;
+
+	version=atoi(GWEN_DB_GroupName(dbV));
+	if (version>0) {
+	  GWEN_DB_NODE *dbT;
+
+	  dbT=GWEN_DB_FindFirstGroup(dbV, "tanMethod");
+	  if (!dbT) {
+	    DBG_INFO(AQHBCI_LOGDOMAIN, "No tan method found");
+	  }
+	  while(dbT) {
+	    AH_TAN_METHOD *tm;
+	    const char *s;
+
+	    tm=AH_TanMethod_new();
+	    AH_TanMethod_SetFunction(tm, GWEN_DB_GetIntValue(dbT, "function", 0, 0));
+	    AH_TanMethod_SetProcess(tm, GWEN_DB_GetIntValue(dbT, "process", 0, 0));
+	    AH_TanMethod_SetMethodId(tm, GWEN_DB_GetCharValue(dbT, "methodId", 0, 0));
+	    AH_TanMethod_SetMethodName(tm, GWEN_DB_GetCharValue(dbT, "methodName", 0, 0));
+	    AH_TanMethod_SetTanMaxLen(tm, GWEN_DB_GetIntValue(dbT, "tanMaxLen", 0, 0));
+	    AH_TanMethod_SetFormatId(tm, GWEN_DB_GetCharValue(dbT, "formatId", 0, 0));
+	    AH_TanMethod_SetPrompt(tm, GWEN_DB_GetCharValue(dbT, "prompt", 0, 0));
+	    AH_TanMethod_SetReturnMaxLen(tm, GWEN_DB_GetIntValue(dbT, "returnMaxLen", 0, 0));
+	    AH_TanMethod_SetMaxActiveLists(tm, GWEN_DB_GetIntValue(dbT, "maxActiveLists", 0, 0));
+	    AH_TanMethod_SetGvVersion(tm, GWEN_DB_GetIntValue(dbT, "gvVersion", 0, 0));
+	    s=GWEN_DB_GetCharValue(dbT, "multiTanAllowed", 0, NULL);
+	    if (s && strcasecmp(s, "j")==0)
+	      AH_TanMethod_SetMultiTanAllowed(tm, 1);
+	    AH_TanMethod_SetTimeShiftAllowed(tm, GWEN_DB_GetIntValue(dbT, "timeShiftAllowed", 0, 0));
+	    AH_TanMethod_SetTanListMode(tm, GWEN_DB_GetIntValue(dbT, "tanListMode", 0, 0));
+	    s=GWEN_DB_GetCharValue(dbT, "stornoAllowed", 0, NULL);
+	    if (s && strcasecmp(s, "j")==0)
+	      AH_TanMethod_SetStornoAllowed(tm, 1);
+	    s=GWEN_DB_GetCharValue(dbT, "needChallengeClass", 0, NULL);
+	    if (s && strcasecmp(s, "j")==0)
+	      AH_TanMethod_SetNeedChallengeClass(tm, 1);
+	    s=GWEN_DB_GetCharValue(dbT, "needChallengeAmount", 0, NULL);
+	    if (s && strcasecmp(s, "j")==0)
+	      AH_TanMethod_SetNeedChallengeAmount(tm, 1);
+	    AH_TanMethod_SetInitMode(tm, GWEN_DB_GetIntValue(dbT, "initMode", 0, 0));
+	    s=GWEN_DB_GetCharValue(dbT, "tanMediumIdNeeded", 0, NULL);
+	    if (s && strcasecmp(s, "j")==0)
+	      AH_TanMethod_SetNeedTanMediumId(tm, 1);
+	    AH_TanMethod_SetMaxActiveTanMedia(tm, GWEN_DB_GetIntValue(dbT, "maxActiveMedia", 0, 0));
+
+	    DBG_INFO(AQHBCI_LOGDOMAIN,
+		     "Adding TAN method %d [%s] for GV version %d",
+		     AH_TanMethod_GetFunction(tm),
+		     AH_TanMethod_GetMethodId(tm),
+		     version);
+	    AH_TanMethod_SetGvVersion(tm, version);
+	    AH_TanMethod_List_Add(tm, ue->tanMethodDescriptions);
+
+	    dbT=GWEN_DB_FindNextGroup(dbT, "tanMethod");
+	  }
+	}
+
+	dbV=GWEN_DB_GetNextGroup(dbV);
+      }
+    }
+    GWEN_DB_Group_free(db);
   }
 }
 
 
 
+void AH_User_LoadSepaDescriptors(AB_USER *u) {
+  AH_USER *ue;
+  GWEN_DB_NODE *db;
+  int rv;
 
-const AH_TAN_METHOD_LIST *AH_User_GetTanMethodDescriptions(const AB_USER *u) {
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  /* read directly from BPD */
+
+  GWEN_StringList_Clear(ue->sepaDescriptors);
+  db=GWEN_DB_Group_new("bpd");
+  rv=AH_Job_SampleBpdVersions("JobGetAccountSepaInfo", u, db);
+  if (rv<0) {
+      DBG_INFO(AQHBCI_LOGDOMAIN, "No BPD for TAN job");
+  }
+  else {
+    GWEN_DB_NODE *dbV;
+
+    dbV=GWEN_DB_GetFirstGroup(db);
+    while(dbV) {
+      int version;
+
+      version=atoi(GWEN_DB_GroupName(dbV));
+      if (version>0) {
+	GWEN_DB_NODE *dbT;
+
+	/* always overwrite with latest version received */
+	GWEN_StringList_Clear(ue->sepaDescriptors);
+	dbT=GWEN_DB_FindFirstGroup(dbV, "SupportedSepaFormats");
+	if (!dbT) {
+	  DBG_INFO(AQHBCI_LOGDOMAIN, "No SEPA descriptor found");
+	}
+	while(dbT) {
+	  int i;
+
+	  for (i=0; i<100; i++) {
+	    const char *s;
+
+	    s=GWEN_DB_GetCharValue(dbT, "format", i, NULL);
+	    if (! (s && *s))
+	      break;
+	    GWEN_StringList_AppendString(ue->sepaDescriptors, s, 0, 1);
+	    DBG_INFO(AQHBCI_LOGDOMAIN,
+		     "Adding SEPA descriptor [%s] for GV version %d",
+		     s, version);
+	  }
+
+	  dbT=GWEN_DB_FindNextGroup(dbT, "SupportedSepaFormats");
+	}
+      }
+
+      dbV=GWEN_DB_GetNextGroup(dbV);
+    }
+  }
+  GWEN_DB_Group_free(db);
+}
+
+
+
+
+const AH_TAN_METHOD_LIST *AH_User_GetTanMethodDescriptions(AB_USER *u) {
   AH_USER *ue;
 
   assert(u);
   ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
   assert(ue);
 
+  /* always reload TAN methods from BPD */
+  AH_User_LoadTanMethods(u);
+
   return ue->tanMethodDescriptions;
 }
 
 
 
+int AH_User_GetMaxTransfersPerJob(const AB_USER *u){
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  return ue->maxTransfersPerJob;
+}
 
 
 
+void AH_User_SetMaxTransfersPerJob(AB_USER *u, int i){
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  ue->maxTransfersPerJob=i;
+}
 
 
 
+int AH_User_GetMaxDebitNotesPerJob(const AB_USER *u){
+  AH_USER *ue;
 
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  return ue->maxDebitNotesPerJob;
+}
+
+
+
+void AH_User_SetMaxDebitNotesPerJob(AB_USER *u, int i){
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  ue->maxDebitNotesPerJob=i;
+}
+
+
+
+int AH_User_AddTextWithoutTags(const char *s, GWEN_BUFFER *obuf) {
+  while(*s) {
+    if (*s=='<') {
+      const char *s2;
+      int l;
+
+      s2=s;
+      s2++;
+      while(*s2 && *s2!='>')
+	s2++;
+      l=s2-s-2;
+      if (l>0) {
+	const char *s3;
+
+	s3=s;
+	s3++;
+	if (l==2) {
+	  if (strncasecmp(s3, "br", 2)==0)
+	    GWEN_Buffer_AppendString(obuf, "\n");
+	}
+	else if (l==3) {
+	  if (strncasecmp(s3, "br/", 3)==0)
+	    GWEN_Buffer_AppendString(obuf, "\n");
+	}
+      }
+      s=s2; /* set s to position of closing bracket */
+    }
+    else
+      GWEN_Buffer_AppendByte(obuf, *s);
+    /* next char */
+    s++;
+  }
+
+  return 0;
+}
+
+
+
+int AH_User_InputTanWithChallenge2(AB_USER *u,
+				   const char *challenge,
+				   const char *challengeHhd,
+				   char *pwbuffer,
+				   int minLen,
+				   int maxLen){
+  int rv;
+  char buffer[1024];
+  const char *un;
+  const char *bn=NULL;
+  GWEN_BUFFER *nbuf;
+  GWEN_BUFFER *xbuf;
+  AB_BANKINFO *bi;
+  uint32_t iflags=0;
+
+  assert(u);
+  un=AB_User_GetUserId(u);
+  /* find bank name */
+  bi=AB_Banking_GetBankInfo(AB_User_GetBanking(u),
+			    "de",
+                            "*",
+			    AB_User_GetBankCode(u));
+  if (bi)
+    bn=AB_BankInfo_GetBankName(bi);
+  if (!bn)
+    AB_User_GetBankCode(u);
+
+  iflags=GWEN_GUI_INPUT_FLAGS_TAN | GWEN_GUI_INPUT_FLAGS_SHOW;
+
+  buffer[0]=0;
+  buffer[sizeof(buffer)-1]=0;
+
+  xbuf=GWEN_Buffer_new(0, 256, 0, 1);
+
+  if (challengeHhd && *challengeHhd) {
+    DBG_ERROR(AQHBCI_LOGDOMAIN, "ChallengeHHD is [%s]", challengeHhd);
+    GWEN_Buffer_AppendString(xbuf, "$OBEGIN$");
+    rv=AH_HHD14_Translate(challengeHhd, xbuf);
+    if (rv<0) {
+      GWEN_Buffer_free(xbuf);
+      AB_BankInfo_free(bi);
+      return rv;
+    }
+    GWEN_Buffer_AppendString(xbuf, "$OEND$");
+    iflags|=GWEN_GUI_INPUT_FLAGS_OPTICAL;
+
+    /* text version */
+    snprintf(buffer, sizeof(buffer)-1,
+	     I18N("Please enter the TAN\n"
+		  "for user %s at %s.\n"), un, bn);
+    buffer[sizeof(buffer)-1]=0;
+    GWEN_Buffer_AppendString(xbuf, buffer);
+    if (challenge && *challenge) {
+      GWEN_Buffer_AppendString(xbuf, challenge);
+      GWEN_Buffer_AppendString(xbuf, "\n");
+    }
+    else {
+      GWEN_Buffer_AppendString(xbuf, I18N("Please enter the TAN from the device."));
+    }
+  }
+  else if (challenge && *challenge) {
+    const char *s;
+
+    /* look for "CHLGUC" */
+    s=GWEN_Text_StrCaseStr(challenge, "CHLGUC");
+    if (s) {
+      DBG_ERROR(AQHBCI_LOGDOMAIN, "Challenge contains CHLGUC");
+      GWEN_Buffer_AppendString(xbuf, "$OBEGIN$");
+      rv=AH_HHD14_Translate(s, xbuf);
+      if (rv<0) {
+	GWEN_Buffer_free(xbuf);
+	AB_BankInfo_free(bi);
+	return rv;
+      }
+      GWEN_Buffer_AppendString(xbuf, "$OEND$");
+      iflags|=GWEN_GUI_INPUT_FLAGS_OPTICAL;
+
+      /* text version */
+      snprintf(buffer, sizeof(buffer)-1,
+               I18N("Please enter the TAN\n"
+                    "for user %s at %s.\n"), un, bn);
+      buffer[sizeof(buffer)-1]=0;
+      GWEN_Buffer_AppendString(xbuf, buffer);
+
+      /* extract text */
+      s=GWEN_Text_StrCaseStr(challenge, "CHLGTEXT");
+      if (s) {
+	/* skip "CHLGTEXT" and 4 digits */
+	s+=12;
+	/* add rest of the message (replace HTML tags, if any) */
+	AH_User_AddTextWithoutTags(s, xbuf);
+      }
+      else {
+	/* create own text */
+	GWEN_Buffer_AppendString(xbuf, I18N("Please enter the TAN from the device."));
+      }
+    }
+    else {
+      /* no optical challenge */
+      DBG_ERROR(AQHBCI_LOGDOMAIN, "Challenge contains no optical data");
+      GWEN_Buffer_AppendString(xbuf, I18N("The server provided the following challenge:"));
+      GWEN_Buffer_AppendString(xbuf, "\n");
+      GWEN_Buffer_AppendString(xbuf, challenge);
+    }
+  }
+
+  nbuf=GWEN_Buffer_new(0, 256 ,0 ,1);
+  AH_User_MkTanName(u, challenge, nbuf);
+  rv=GWEN_Gui_GetPassword(iflags,
+			  GWEN_Buffer_GetStart(nbuf),
+			  I18N("Enter TAN"),
+			  GWEN_Buffer_GetStart(xbuf),
+			  pwbuffer,
+			  minLen,
+			  maxLen,
+			  0);
+  GWEN_Buffer_free(xbuf);
+  GWEN_Buffer_free(nbuf);
+  AB_BankInfo_free(bi);
+  return rv;
+}
+
+
+
+const char *AH_User_FindSepaDescriptor(AB_USER *u, const char *tmpl) {
+  AH_USER *ue;
+  GWEN_STRINGLISTENTRY *se;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  if (GWEN_StringList_Count(ue->sepaDescriptors)<1)
+    AH_User_LoadSepaDescriptors(u);
+
+  se=GWEN_StringList_FirstEntry(ue->sepaDescriptors);
+  while(se) {
+    const char *s;
+
+    s=GWEN_StringListEntry_Data(se);
+    if (s && *s && -1!=GWEN_Text_ComparePattern(s, tmpl, 1))
+      return s;
+
+    se=GWEN_StringListEntry_Next(se);
+  }
+
+  return NULL;
+}
+
+
+
+const GWEN_STRINGLIST *AH_User_GetSepaDescriptors(AB_USER *u) {
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  return ue->sepaDescriptors;
+}
+
+
+
+const char *AH_User_GetSepaTransferProfile(const AB_USER *u) {
+  const AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  return ue->sepaTransferProfile;
+}
+
+
+
+void AH_User_SetSepaTransferProfile(AB_USER *u, const char *profileName) {
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  free(ue->sepaTransferProfile);
+  if (profileName)
+    ue->sepaTransferProfile=strdup(profileName);
+  else
+    ue->sepaTransferProfile=NULL;
+}
+
+
+
+const char *AH_User_GetSepaDebitNoteProfile(const AB_USER *u) {
+  const AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  return ue->sepaDebitNoteProfile;
+}
+
+
+
+void AH_User_SetSepaDebitNoteProfile(AB_USER *u, const char *profileName) {
+  AH_USER *ue;
+
+  assert(u);
+  ue=GWEN_INHERIT_GETDATA(AB_USER, AH_USER, u);
+  assert(ue);
+
+  free(ue->sepaDebitNoteProfile);
+  if (profileName)
+    ue->sepaDebitNoteProfile=strdup(profileName);
+  else
+    ue->sepaDebitNoteProfile=NULL;
+}
